"""API constants

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Limit pykern imports
from pykern.pkcollections import PKDict
import datetime
import enum
import inspect
import msgpack
import pykern.util

#: API that authenticates connections (needed for client)
AUTH_API_NAME = "authenticate_connection"

#: API version for AUTH (and for pykern.api)
AUTH_API_VERSION = 658584001

# Protocol code shared between client & server, not public

# A bit of type checking
_MSG_KIND_BASE = 777500


class MsgKind(enum.Enum):
    CALL = _MSG_KIND_BASE + 1
    REPLY = _MSG_KIND_BASE + 2
    SUBSCRIBE = _MSG_KIND_BASE + 3
    UNSUBSCRIBE = _MSG_KIND_BASE + 4

    def is_call(self):
        return self is self.CALL

    def is_reply(self):
        return self is self.REPLY

    def is_subscribe(self):
        return self is self.SUBSCRIBE

    def is_unsubscribe(self):
        return self is self.UNSUBSCRIBE


_MSG_KIND_IS_VALID = PKDict(
    client=frozenset((MsgKind.REPLY, MsgKind.UNSUBSCRIBE)),
    server=frozenset((MsgKind.CALL, MsgKind.SUBSCRIBE, MsgKind.UNSUBSCRIBE)),
)


_SUBSCRIPTION_ATTR = "pykern_api_util_subscription"


class APICallError(pykern.util.APIError):
    """Raised when call execution ends in exception or other error"""

    def __init__(self, error):
        super().__init__("error={}", error)


class APIDisconnected(pykern.util.APIError):
    """Raised when remote server closed or other error"""

    def __init__(self):
        super().__init__("")


class APIForbidden(pykern.util.APIError):
    """Raised for forbidden or protocol error"""

    def __init__(self):
        super().__init__("")


class APIKindError(pykern.util.APIError):
    """Raised when kind mismatch"""

    def __init__(self, error):
        super().__init__("error={}", error)


class APINotFound(pykern.util.APIError):
    """Raised for an object not found"""

    def __init__(self, api_name):
        super().__init__("api_name={}", api_name)


class APIProtocolError(pykern.util.APIError):
    """Raised when protocol error at lower level"""

    def __init__(self, error):
        super().__init__("error={}", error)


def is_subscription(func):
    """Is `func` a subscription api?

    Args:
        func (function): class api
    Returns:
        bool: True if is subscription api
    """
    return getattr(func, _SUBSCRIPTION_ATTR, False)


def msg_pack(unserialized):
    """Used by client and server, not public"""

    def _default(obj):
        if isinstance(obj, datetime.datetime):
            return int(obj.timestamp())
        if isinstance(obj, enum.Enum):
            return obj.value
        if hasattr(obj, "tolist"):
            # tolist works with pandas and numpy. If tolist takes
            # params or not a callable, then the result will be the
            # essentially the same as not having this code.
            return obj.tolist()
        return obj

    p = msgpack.Packer(autoreset=False, default=_default)
    p.pack(unserialized)
    # TODO(robnagler) getbuffer() would be better
    return p.bytes()


def msg_unpack(serialized, which):
    """Used by client and server, not public"""

    def _int(rv, which):
        i = rv[which]
        if not isinstance(i, int):
            return None, f"msg {which} non-integer type={type(i)}"
        if i <= 0:
            return None, f"msg {which} non-positive int={i}"
        return None

    def _kind(rv):
        if r := _int(rv, "msg_kind"):
            return r
        try:
            k = MsgKind(rv.msg_kind)
            if k not in _MSG_KIND_IS_VALID[which]:
                return None, f"{k} invalid for {which}"
            rv.msg_kind = k
            return None
        except Exception as e:
            return None, f"msg_kind={rv.msg_kind} not in valid"

    try:
        u = msgpack.Unpacker(
            object_hook=pykern.pkcollections.object_pairs_hook,
        )
        u.feed(serialized)
        rv = u.unpack()
    except Exception as e:
        return None, f"msgpack exception={e}"
    if not isinstance(rv, PKDict):
        return None, f"msg not dict type={type(rv)}"
    for f in "call_id", "msg_kind":
        if not rv.get(f):
            return None, f"msg missing {f} keys={list(rv.keys())}"
    return _int(rv, "call_id") or _kind(rv) or (rv, None)


def subscription(func):
    """Decorator for api functions thhat can be subscribed by clients.

    Args:
        func (function): class api
    Returns:
        function: function to use
    """

    # Give some early feedback
    if not inspect.iscoroutinefunction(func):
        raise AssertionError(f"func={func.__name__} must be a coroutine")
    setattr(func, _SUBSCRIPTION_ATTR, True)
    return func



# -*- coding: utf-8 -*-
"""Unit test for `pykern.pksubprocess`

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import os
import pytest
import shutil


@pytest.mark.skipif(
    not shutil.which("setsid"),
    reason="no setsid command",
)
def test_check_call_with_signals():
    from pykern import pksubprocess
    from pykern import pkunit, pkcompat, pkdebug
    import os
    import signal
    import subprocess
    import time

    messages = []

    def msg(*args):
        s = args[0]
        messages.append(s.format(*args[1:]))

    signals = []

    def signal_handler(sig, frame):
        signals.append(sig)

    with pkunit.save_chdir_work():
        with open("true.out", "w+") as o:
            pksubprocess.check_call_with_signals(["true"], output=o)
            o.seek(0)
            actual = o.read()
            assert "" == actual, 'Expecting empty output "{}"'.format(actual)

        with open("echo.out", "w+") as o:
            messages = []
            tag = "xyzzy"
            pksubprocess.check_call_with_signals(["echo", tag], output=o, msg=msg)
            o.seek(0)
            actual = o.read()
            assert tag in actual, '"{}" not in output "{}"'.format(tag, actual)
            assert "started" in messages[0], '"started" not in messages[0] "{}"'.format(
                messages[0]
            )
            assert (
                "normal exit" in messages[1]
            ), '"normal exit" not in messages[1] "{}"'.format(messages[1])

        with open("kill.out", "w+") as o:
            messages = []
            signals = []
            signal.signal(signal.SIGTERM, signal_handler)
            with open("kill.sh", "w") as f:
                # Need to wait before sending signal, because subprocess needs to allow
                # this process (parent) to run
                f.write(
                    f"""
sleep .2
kill -TERM {os.getpid()}
sleep 5
echo FAIL
"""
                )
            cmd = ["sh", "kill.sh"]
            exc = None
            try:
                pksubprocess.check_call_with_signals(cmd, output=o, msg=msg)
            except RuntimeError as e:
                exc = e
            except BaseException as e:
                pkunit.pkfail("unexpected exception={}", e)
            o.seek(0)
            actual = o.read()
            pkunit.pkeq("", actual)
            pkunit.pkok(
                signal.SIGTERM in signals, '"SIGTERM" not in signals "{}"', signals
            )
            pkunit.pkre("error exit", messages[1])
            if exc is None:
                pkunit.pkfail("exception was not raised")

        with open("kill.out", "w+") as o:
            messages = []
            signals = []
            signal.signal(signal.SIGTERM, signal_handler)
            with open("kill.sh", "w") as f:
                f.write(
                    f"""
setsid bash -c "sleep 1; echo FAIL; setsid sleep 1313 & disown" &
disown
sleep .2
kill -TERM {os.getpid()}
sleep 5
echo FAIL
"""
                )
            cmd = ["bash", "kill.sh"]
            exc = None
            try:
                pksubprocess.check_call_with_signals(
                    cmd,
                    output=o,
                    msg=msg,
                    recursive_kill=True,
                )
            except RuntimeError as e:
                exc = e
            except BaseException as e:
                pkunit.pkfail("unexpected exception={}", e)
            time.sleep(2)
            o.seek(0)
            actual = o.read()
            assert "" == actual, 'Expecting empty output "{}"'.format(actual)
            assert signal.SIGTERM in signals, '"SIGTERM" not in signals "{}"'.format(
                signals
            )
            assert (
                "error exit" in messages[1]
            ), '"error exit" not in messages[1] "{}"'.format(messages[1])
            p = pkcompat.from_bytes(
                subprocess.check_output(
                    ["ps", "axfj"],
                    stdin=open(os.devnull),
                    stderr=subprocess.STDOUT,
                ),
            )
            assert "sleep 1313" not in p, "sleep did not terminate: {}".format(p)



# -*- coding: utf-8 -*-
"""Wrapper for Python's :mod:`platform` to provide cleaner programmatic
control of system features.

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

import platform
import re
import sys


def is_darwin():
    """All flavors of Mac (OS X)

    Returns:
        bool: True if :attr:`sys.platform` is Mac.
    """
    return sys.platform.startswith("darwin")


def is_linux():
    """All flavors of linux

    Returns:
        bool: True if :attr:`sys.platform` is Linux.
    """
    return sys.platform.startswith("linux")


def is_unix():
    """All flavors of Unix. Currently, it's darwin, linux, cygwin, freebsd,
    netbsd, sunos, solaris, unixware, irix, aix, and next.

    Returns:
        bool: True if :attr:`sys.platform` is a pure Unix system (e.g. not beos)
    """
    return (
        re.match(
            r"aix|cygwin|darwin|freebsd|irix|linux|netbsd|solaris|sunos|unix",
            sys.platform,
        )
        is not None
    )


def is_windows():
    """All flavors of Windows (32, 64, cygwin, etc.). If your program expects
    a unix flavor, you will want :func:`is_unix`.

    Returns:
        bool: True if :attr:`sys.platform` is ``win32`` or ``cygwin``
    """
    return re.match(r"win|cygwin", sys.platform) is not None



"""PyTest for :mod:`pykern.util`

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_dev_run_dir():
    from pykern import util, pkunit

    pkunit.pkeq(pkunit.work_dir(), util.dev_run_dir(util))
    util._dev_run_dir = None
    p = pkunit.is_test_run
    try:
        pkunit.is_test_run = lambda: False
        pkunit.pkeq(
            pkunit.work_dir().dirpath().dirpath().join("run"),
            util.dev_run_dir(util),
        )
    finally:
        pkunit.is_test_run = p


def test_is_pure_text():
    from pykern import util
    from pykern import pkunit

    def _false(value, is_truncated=False):
        pkunit.pkeq(False, util.is_pure_text(value, is_truncated=is_truncated))

    def _true(value, is_truncated=False):
        pkunit.pkeq(True, util.is_pure_text(value, is_truncated=is_truncated))

    a = "a".encode("utf-8")
    _false(b"\0 one null causes failure in valid text")
    _false(a + b"\xc2")
    _false(bytes(range(1, 0x20)))
    _false(b"\xd4\x16\xc0\xd6\xec\xbf\x92\xe6\x84T\xc9 \xe9\xbf")
    # backwards probing on non-text case
    _false(a + b"\xc2\xc2\xc2\xc2", is_truncated=True)
    # boundary of control code ratio
    _false(b"\x01" * 33 + b"\x07" * 67)
    _true(b"\x01" * 32 + b"\x07" * 68)
    _true(b"")
    _true(b"This is example text")
    _true(b"\x07\x08\t\n\x0b\x0c\r\x0e\x0f")
    # backwards probing on text case
    _true(a + "¡".encode("utf-8"), is_truncated=True)
    _true(a + b"\xf0\x9f\x8c\xae", is_truncated=True)



# -*- coding: utf-8 -*-
"""pkconfig init testing

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_init(pkconfig_setup):
    """Validate parse_tuple"""
    pkconfig = pkconfig_setup(
        cfg=dict(P1_M1_TUPLE3="", P1_M1_TUPLE4="aaa:bbbb"),
        env=dict(P1_M1_REQ8="99"),
    )
    from p1.m1 import cfg

    assert () == cfg.tuple1, "When tuple1 is none, is empty"
    assert (1,) == cfg.tuple2, "When tuple2 is none, is (1,)"
    pkconfig.reset_state_for_testing()
    assert () == cfg.tuple3, "tuple3 should be overriden to be empty"
    assert (
        "aaa",
        "bbbb",
    ) == cfg.tuple4, 'tuple4 should be overriden to be ("aaa", "bbbb")'
    assert ("dev-host",) == cfg.req10



"""xlsx_test case

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
import pykern.xlsx

PATH = "case2.xlsx"
w = pykern.xlsx.Workbook(path=PATH)
s = w.sheet(title="s1")
t = s.table(title="t1", defaults=PKDict(round_digits=0, num_fmt="currency"))
t.header(
    name="Name",
    count="Count",
    other="Other",
    row_sum="RowSum",
)
for i in range(1, 5):
    t.row(
        name=f"c{i}",
        count=t.cell(i, link=["count", f"row{i}"]),
        # Tests a bug link references
        other=t.cell(i, link=[f"row{i}", "other"]),
        row_sum=t.cell(["+", [f"row{i}"]], link="row_sum"),
    )
t.footer(
    name="Product",
    count=t.cell(["*", ["count"]], link="prod"),
    other=None,
    row_sum=t.cell(["*", ["row_sum"]]),
)
s = w.sheet(title="s2")
t = s.table(title="t2", defaults=PKDict(round_digits=0, num_fmt="currency"))
t.row(
    prod=["prod"],
    prod_number=t.cell(["prod"], fmt="text"),
)
w.save()



# -*- coding: utf-8 -*-
"""PyTest for :mod:`pykern.pkinspect`

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_append_exception_reason():
    from pykern import pkinspect
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkeq

    # Run through the cases because there are so many
    r = "xyzzy"
    for e, a in (
        (PKDict(), PKDict()),
        (PKDict(reason=1), PKDict(reason=1)),
        (PKDict(reason="xyzzy"), PKDict(reason="")),
        (PKDict(args=("xyzzy",)), PKDict(args=None)),
        (PKDict(args=("xyzzy",)), PKDict(args=())),
        (PKDict(args=(1,)), PKDict(args=(1,))),
        (PKDict(args=("hello; xyzzy",)), PKDict(args=("hello",))),
    ):
        pkinspect.append_exception_reason(a, r)
        pkeq(e, a)


def test_module_basename():
    from pykern import pkinspect
    from pykern import pkunit

    p1 = pkunit.import_module_from_data_dir("p1")
    assert pkinspect.module_basename(p1) == "p1"
    m1 = pkunit.import_module_from_data_dir("p1.m1")
    assert pkinspect.module_basename(m1) == "m1"
    assert pkinspect.module_basename(m1.C) == "m1"
    assert pkinspect.module_basename(m1.C()) == "m1"
    assert pkinspect.module_basename(m1) == "m1"
    assert pkinspect.module_basename(m1.C) == "m1"
    assert pkinspect.module_basename(m1.C()) == "m1"
    p2 = pkunit.import_module_from_data_dir("p1.p2")
    assert pkinspect.module_basename(p2) == "p2"
    m2 = pkunit.import_module_from_data_dir("p1.p2.m2")
    assert pkinspect.module_basename(m2) == "m2"


def test_caller():
    from pykern import pkinspect
    from pykern import pkunit
    import inspect
    import sys

    def _func_name(expect):
        pkunit.pkeq(expect, pkinspect.caller_func_name())

    m1 = pkunit.import_module_from_data_dir("p1.m1")
    c = m1.caller()
    expect = inspect.currentframe().f_lineno - 1
    pkunit.pkeq(expect, c.lineno)
    expect = "test_caller"
    pkunit.pkeq(expect, c.name)
    _func_name(expect)
    this_module = sys.modules[__name__]
    c = m1.caller(ignore_modules=[this_module])
    pkunit.pkne(expect, c.name)
    my_caller = pkinspect.caller()
    expect = my_caller._module
    pkunit.pkeq(expect, c._module)
    pkunit.pkeq(expect, c._module)


def test_caller_module():
    from pykern import pkunit

    m1 = pkunit.import_module_from_data_dir("p1.m1")
    assert (
        __name__ == m1.caller_module().__name__
    ), "caller_module should return this module"
    n = m1.caller_module(exclude_first=False).__name__
    expect = m1.__name__
    assert expect == n, "{}: should be {}".format(n, expect)


def test_is_caller_main():
    import sys
    import subprocess
    from pykern import pkio
    from pykern import pkunit

    m1 = pkunit.import_module_from_data_dir("p1.m1")
    assert not m1.is_caller_main(), "When not called from main, is_caller_main is False"
    with pkio.save_chdir(pkunit.data_dir()):
        subprocess.check_call(
            [sys.executable, "-c", "from p1 import m1; assert m1.is_caller_main()"]
        )


def test_is_valid_identifier():
    from pykern import pkinspect

    assert pkinspect.is_valid_identifier("_"), "a single underscore is valid"
    assert pkinspect.is_valid_identifier("A_3"), "any letters and numbers is valid"
    assert not pkinspect.is_valid_identifier("1abc"), "a leading number is invalid"
    assert not pkinspect.is_valid_identifier(""), "empty string is invalid"


def test_submodule_name():
    from pykern import pkinspect
    from pykern import pkunit

    m2 = pkunit.import_module_from_data_dir("p1.p2.m2")
    assert pkinspect.submodule_name(m2) == "p2.m2"


def test_root_pkg():
    from pykern import pkinspect
    from pykern import pkunit

    m2 = pkunit.import_module_from_data_dir("p1.p2.m2")
    assert pkinspect.root_package(m2) == "p1"



"""Excel spreadsheet generator

:copyright: Copyright (c) 2021 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp, pkdformat
import decimal
import xlsxwriter
import re


_XL_COLS = None

_NO_CHILDREN = tuple()

_DEFAULT_ROUND_DIGITS = 2

_DIGITS_TO_PLACES = None

_SPACES = re.compile(r"\s+")

#: For sanity sake, we start at 1 with column numbers
_COL_NUM_1 = 1

_ROW_NUM_1 = 1

#: max rows is 1048576 so just use 10M. Used for sort_index
_ROW_MODULUS = 10000000

# max range of operands for SUM, AND, etc.
_MULTI_COUNT = (1, 65535)


class _SimpleBase(PKDict):
    def __str__(self):
        return self.pkdebug_str()

    def pkdebug_str(self):
        l = ""
        for x in "link", "value", "xl_id":
            l += f",{x}={self[x]}" if x in self else ""
        for x in "content", "title", "path", "token", "_content":
            if x in self:
                return f"{self.__class__.__name__}({x}={self[x]}{l})"
        return f"{self.__class__.__name__}({l})"

    def _error(self, fmt, *args, **kwargs):
        # TODO: print stack
        raise AssertionError(pkdformat(fmt + "; {self}", *args, self=self, **kwargs))


class _Base(_SimpleBase):
    def __init__(self, cfg):
        super().__init__()
        self.pkupdate(cfg).pksetdefault(defaults=PKDict)

    def cell(self, content_or_cell, **kwargs):
        """Convert content or a cell config to a `_Cell`

        If `content_or_cell` is a PKDict, is used to configure a cell. `kwargs` must be empty.

        If `content_or_cell` is a `_Cell`, returns itself. `kwargs` must be empty.

        Otherwise, `content_or_cell` is treated as content and `_Cell` is created with the content and `kwargs`.

        Args:
            content_or_cell (object): see above

        Returns:
            _Cell: instance
        """
        if isinstance(content_or_cell, _Cell):
            assert not kwargs
            return content_or_cell
        elif isinstance(content_or_cell, dict):
            assert not kwargs
            return _Cell(content_or_cell)
        else:
            kwargs["content"] = content_or_cell
            return _Cell(kwargs)

    def _cascade_defaults(self, parent_defaults):
        self.defaults.pksetdefault(**parent_defaults)
        for c in self._children():
            c._cascade_defaults(self.defaults)
        # TODO(robnagler): document clearing defaults with None
        for k in [k for k, v in self.defaults.items() if v is None]:
            del self.defaults[k]

    def _child(self, children, child, kwargs):
        s = child(kwargs)
        s._relations(self)
        children.append(s)
        return s

    def _compile_pass1(self):
        for c in self._children():
            c._compile_pass1()

    def _compile_pass2(self):
        for c in self._children():
            c._compile_pass2()

    def _relations(self, parent):
        self.parent = parent
        if isinstance(parent, _Sheet):
            self.sheet = parent
            self.workbook = parent.workbook
        elif isinstance(parent, Workbook):
            self.workbook = parent
        else:
            self.sheet = parent.sheet
            self.workbook = parent.workbook
        return self

    def _print(self):
        pkdlog(self)
        for c in self._children():
            c._print()


class Workbook(_Base):
    def __init__(self, **kwargs):
        """Creates Workbook

        Args:
            path (py.path): where to write the spreadsheet
            defaults (PKDict): default values, e.g. round_digits
        """
        super().__init__(kwargs)
        self.sheets = []
        # for consistency in error messages
        self.workbook = self
        self.links = PKDict()

    def xl_fmt(self, cfg):
        """Get the Excel format for cfg

        Args:
            cfg (PKDict): key values that are supported by xlsxwriter

        Returns:
            Format: object which represents format
        """
        k = str(cfg)
        return self._xl_fmt.pksetdefault(k, lambda: self.xl.add_format(cfg))[k]

    def sheet(self, **kwargs):
        """Append a sheet to a Workbook

        Args:
            title (str): label for the sheet
            defaults (PKDict): default values, e.g. round_digits
        """
        return self._child(self.sheets, _Sheet, kwargs)

    def save(self):
        try:
            self._cascade_defaults(PKDict(str_fmt="text", num_fmt="decimal"))
            self._compile_pass1()
            self._compile_pass2()
            self._xl_fmt = PKDict()
            self.xl = xlsxwriter.Workbook(str(self.path))
            for s in self.sheets:
                s._save()
            self.xl.close()
            self._xl_fmt = None
            self.xl = None
        except Exception:
            pkdlog("workbook save failed; path={}", self.workbook.path)
            raise

    def _assert_link_pair(self, link, left, right):
        e = None
        for x in (
            ["fmt", left.get("fmt"), right.get("fmt")],
            ["is_decimal", left.is_decimal, right.is_decimal],
            ["type", type(left.expr.py_value()), type(right.expr.py_value())],
            ["round_digits", left.round_digits, right.round_digits],
        ):
            if x[1] != x[2]:
                self._error(
                    "link={} {} {}={} different from {} {}={}",
                    link,
                    left,
                    x[0],
                    x[1],
                    right,
                    x[0],
                    x[2],
                )

    def _children(self):
        return self.sheets

    def _compile_pass2(self):
        super()._compile_pass2()
        for k, v in self.links.items():
            for i in range(len(v) - 1):
                self._assert_link_pair(k, v[i], v[i + 1])


class _Cell(_Base):
    def compile_pass2_for_link_ref(self):
        self._compile_pass2()

    def _children(self):
        return _NO_CHILDREN

    def _compile_pass1(self):
        self._compile_link1()

    def _compile_pass2(self):
        if "is_compiled" in self:
            if not self.is_compiled:
                self._error("circular referenced cell={}", self)
            return
        self.is_compiled = False
        self.expr = _Expr(self.content, self)
        # expression's value overrides defaults
        for k in "fmt", "round_digits":
            if (v := self.expr.get(k)) is not None:
                self.pksetdefault(k, v)
        self.pksetdefault(
            round_digits=lambda: self.defaults.get(
                "round_digits", _DEFAULT_ROUND_DIGITS
            ),
        )
        if (v := self.defaults.get("fmt")) is not None:
            self.pksetdefault(fmt=v)
        self.is_decimal = self.expr.is_decimal()
        self.is_compiled = True

    def _compile_link1(self):
        if "link" in self:
            for l in [self.link] if isinstance(self.link, str) else self.link:
                if not l[0].isalnum():
                    self._error("link={} must begin with alphanumeric", l)
                self.workbook.links.setdefault(l, []).append(self)

    def _save(self):
        f = _Fmt(self)
        self.sheet.width(_XL_COLS[self.col_num], f.width(self))
        f = self.workbook.xl_fmt(f)
        e = self.expr
        if e.is_formula:
            self.sheet.xl.write_formula(
                self.xl_id,
                formula=e.xl_formula_for_cell(self),
                value=e.xl_value_for_cell(self),
                cell_format=f,
            )
        else:
            self.sheet.xl.write(self.xl_id, e.xl_value_for_cell(self), f)
        if self.is_decimal:
            self.sheet.text_cells.append(self.xl_id)


class _Expr(_SimpleBase):
    def __init__(self, content, cell):
        def _formula():
            self.cell = cell
            self.is_formula = True
            self._is_decimal = False
            if len(content) == 0:
                self._error("empty expr")
            if not (self._operator(content) or self._link_ref(content)):
                self._error("invalid op or link as first element of expr={}", content)

        def _literal():
            nonlocal content

            self.resolved_count = 1
            self.is_formula = False
            if content is None:
                content = ""
            elif isinstance(content, (bool, str, decimal.Decimal)):
                # must be here, because bool is a subclass of int
                pass
            elif isinstance(content, (float, int)):
                content = decimal.Decimal(content)
            else:
                self._error(
                    "invalid literal type={} content={}", type(content), content
                )
            self.pksetdefault(
                _py_value=content,
                _is_decimal=isinstance(content, decimal.Decimal),
            )

        pkdc("{}!{}={}", cell.sheet.title, cell.xl_id, content)
        self.is_formula = isinstance(content, (list, tuple))
        if self.is_formula:
            _formula()
        else:
            _literal()
        self._content = content

    def is_decimal(self):
        return self._is_decimal or isinstance(self.py_value(), decimal.Decimal)

    def py_value(self):
        if (rv := self.get("_py_value")) is None:
            self._error("expected an expression with a single value expr={}", self)
        if callable(rv):
            self._py_value = rv = rv()
        if not isinstance(rv, (bool, str, decimal.Decimal)):
            self._error("unexpected expression value={} expr={}", rv, self)
        return rv

    def xl_formula(self):
        if self.is_formula:
            return self._xl_formula
        rv = self.py_value()
        if isinstance(rv, bool):
            return "TRUE" if rv else "FALSE"
        elif isinstance(rv, str):
            return '"' + rv.replace('"', '""') + '"'
        elif isinstance(rv, decimal.Decimal):
            return str(rv)
        raise AssertionError(pkdformat("unexpected value={} expr={}", rv, self))

    def xl_formula_for_cell(self, cell):
        if self.resolved_count > 1:
            raise self._error(
                "cannot render link with operator link={} cell={}", self, cell
            )
        rv = self._xl_formula
        if self.is_decimal():
            rv = f"ROUND({rv},{cell.round_digits})"
        return f"={rv}"

    def xl_value_for_cell(self, cell):
        if self.resolved_count > 1:
            raise self._error(
                "cannot render link with operator link={} cell={}", self, cell
            )
        rv = self.py_value()
        if isinstance(rv, decimal.Decimal):
            return _rnd(rv, cell.round_digits)
        if isinstance(rv, bool):
            return "TRUE" if rv else "FALSE"
        return rv

    def _link_ref(self, content):
        def _compile():
            l = self.cell.workbook.links
            if content[0] not in l:
                self._error("link={} not found cell={}", content, self.cell)
            rv = PKDict(
                cells=l[content[0]],
                last_cell=None,
                pairs=[],
                resolved_count=0,
            )
            for c in sorted(rv.cells, key=lambda x: x.sort_index):
                c.compile_pass2_for_link_ref()
                rv.resolved_count += 1
                # _ROW_MODULUS ensures a gap so columns are separated by more than "+1"
                # and this will never link the wrong column
                if (
                    rv.last_cell is not None
                    and rv.last_cell.sheet == c.sheet
                    and rv.last_cell.sort_index + 1 == c.sort_index
                ):
                    rv.last_cell = rv.pairs[-1].last = c
                    continue
                rv.pairs.append(PKDict(first=c, last=None))
                rv.last_cell = c
            return rv

        def _xl_formula(pairs):
            rv = ""
            for p in pairs:
                if len(rv):
                    rv += ","
                rv += _xl_id(p.first)
                if p.last is not None:
                    rv += ":" + _xl_id(p.last)
            return rv

        def _xl_id(cell):
            r = ""
            if cell.sheet != self.cell.sheet:
                r = f"'{cell.sheet.title}'!"
            return r + cell.xl_id

        if not (
            len(content) == 1
            and isinstance(content[0], str)
            and content[0][0].isalnum()
        ):
            return False
        c = _compile()
        # _assert_link_pair validates that the cells are the same type, fmt, round_digits
        l = c.last_cell
        self.pkupdate(
            _link_ref=content,
            _is_decimal=l.expr.is_decimal(),
            _py_value=(lambda: l.expr.py_value()) if c.resolved_count == 1 else None,
            _xl_formula=_xl_formula(c.pairs),
            cells=c.cells,
            resolved_count=c.resolved_count,
        )
        self._options([l])
        return True

    def _operator(self, content):
        if not (o := _OpSpec.find(content[0])):
            return False
        self.update(o.evaluate(content[1:], self))
        self._options(self._operands.exprs)
        return True

    # TODO(robnagler) doesn't work with "if" or bool ops
    # Needs to be more logical. Works for links, because they are always consistent
    # operators should cascade values
    def _options(self, exprs_or_cells):
        for k in "fmt", "round_digits":
            if self.get(k) is not None:
                continue
            for x in exprs_or_cells:
                if (v := x.get(k)) is None:
                    continue
                if k not in self:
                    self[k] = v
                # TODO doc that grabs the leftmost fmt or round, or throws away
                elif self[k] is not None and self[k] != v:
                    del self[k]
                    break


class _Fmt(PKDict):
    _MAP = PKDict(
        top=PKDict(top=True),
        bold=PKDict(bold=True),
        currency="$#,##0.0",
        decimal="0.0",
        percent="0.0%",
        text="@",
    )
    _SPECIAL = re.compile("[$%]")
    _WIDTH_SLOP = 1
    _WIDTH_BOOL = max(len("TRUE"), len("FALSE"))

    def width(self, cell):
        v = cell.expr.py_value()
        if not cell.is_decimal:
            return self._WIDTH_BOOL if isinstance(v, bool) else len(v)
        n = self._WIDTH_SLOP
        x = v.as_tuple()
        i = len(x.digits) + x.exponent
        n += i
        if "," in self.num_format:
            n += i // 3
        n += cell.round_digits
        if n and self._SPECIAL.search(self.num_format):
            n += 1
        return n

    def __init__(self, cell):
        def _num(name, attr, digits):
            if digits is None:
                return attr
            if name == "percent":
                digits -= 2
            return attr.replace(
                ".0",
                "" if digits <= 0 else "." + "0" * digits,
            )

        for k in (
            "font",
            "border",
        ):
            if k in cell.defaults:
                self.update(self._MAP[cell.defaults[k]])
        for x in _SPACES.split(cell.get("fmt", "")):
            if len(x) == 0:
                continue
            f = self._MAP[x]
            if isinstance(f, dict):
                self.update(f)
            else:
                self.num_format = _num(x, f, cell.round_digits)
        if "num_format" not in self:
            h = cell.defaults["num_fmt" if cell.is_decimal else "str_fmt"]
            self.num_format = _num(h, self._MAP[h], cell.round_digits)

    def __str__(self):
        return ";".join([f"{k}={self[k]}" for k in sorted(self)])


class _OpSpec(_SimpleBase):
    _instances = PKDict()

    def __init__(self, token, func, **kwargs):
        if token in self._instances:
            raise AssertionError(f"duplicate token={token}")
        super().__init__(kwargs)
        self.pksetdefault(
            _infix=False,
            _is_multi=False,
            operand_count=(2, 2),
            token=token,
        )
        if isinstance(func, str):
            self._py_func = getattr(self, f"_py_func_{func}")
        else:
            self._py_func = self._py_func_binary
            self._py_func_binary_func = func
        if isinstance((f := self.get("_xl_func")), str):
            if x := getattr(self, f"_xl_func_{f}", None):
                self._xl_func = x
            else:
                self._xl_func = self._xl_func_default
                self._xl_func_default_name = f
        elif self._infix:
            self._xl_func = self._xl_func_infix
        else:
            self._xl_func = self._xl_func_default
            self._xl_func_default_name = token
        if isinstance(self._is_decimal, str):
            self._is_decimal = getattr(self, f"_is_decimal_{self._is_decimal}")
        self._instances[token] = self

    @classmethod
    def init(cls):

        def _compare(token, func):
            return cls(token, func, _is_decimal=False, _infix=True)

        def _multi(token, func, init=None, xl=None):
            return cls(
                token,
                "multi",
                _py_func_multi_func=func,
                _py_func_multi_init=(
                    decimal.Decimal(init) if isinstance(init, int) else init
                ),
                _xl_func=xl or token,
                operand_count=_MULTI_COUNT,
                _is_multi=True,
                _is_decimal=True,
            )

        cls("%", lambda x, y: x % y, _xl_func="MOD", _is_decimal=True)
        _multi("*", lambda rv, y: rv * y, 1, "PRODUCT")
        _multi("+", lambda rv, y: rv + y, 0, "SUM")
        _multi("MAX", max)
        _multi("MIN", min)
        cls(
            "-",
            "minus",
            operand_count=(1, 2),
            _is_decimal=True,
            _infix=True,
            _xl_func="minus",
        )
        cls("/", lambda x, y: x / y, _is_decimal=True, _infix=True)
        _compare("<", lambda x, y: x < y)
        _compare("<=", lambda x, y: x <= y)
        _compare("==", lambda x, y: x == y)
        _compare(">", lambda x, y: x > y)
        _compare(">=", lambda x, y: x >= y)
        cls("AND", "and", operand_count=_MULTI_COUNT, _is_decimal=False)
        cls("OR", "or", operand_count=_MULTI_COUNT, _is_decimal=False)
        cls("NOT", "not", operand_count=(1, 1), _is_decimal=False)
        cls("IF", "if", operand_count=(2, 3), _is_decimal="if")

    @classmethod
    def find(cls, token):
        return cls._instances.get(token)

    def evaluate(self, operands, cell):
        o = self._operands(operands, cell)
        return PKDict(
            _is_decimal=(
                self._is_decimal
                if isinstance(self._is_decimal, bool)
                else self._is_decimal(o)
            ),
            _op_spec=self,
            _operands=o,
            _py_value=lambda: self._py_func(o),
            _xl_formula=self._xl_func(o),
            resolved_count=1,
        )

    def _is_decimal_if(self, operands):
        if len(operands.exprs) != 3:
            return False
        return operands.exprs[0].is_decimal() and operands.exprs[1].is_decimal()

    def _operands(self, operands, expr):
        def _count_ok(rv):
            if self._is_multi:
                return self.operand_count[0] <= rv.count <= self.operand_count[1]
            return (
                self.operand_count[0] <= len(rv.exprs) <= self.operand_count[1]
                and len(rv.exprs) == rv.count
            )

        rv = PKDict(count=0, exprs=[])
        for o in operands:
            e = _Expr(o, expr.cell)
            rv.exprs.append(e)
            rv.count += e.resolved_count
        if not _count_ok(rv):
            self._operands_error(rv)
        return rv

    def _operands_error(self, operands):
        x = (
            "; You might need to be more specific link names to avoid automatic link operand grouping."
            # TODO(robnagler) not quite right
            if max(self.operand_count[1], 2) < operands.count
            else ""
        )
        self._error(
            "invalid operand count={} operand_count={} operator={} operands={}{}",
            operands.count,
            self.operand_count,
            self,
            operands.exprs,
            x,
        )

    def _py_func_and(self, operands):
        for e in operands.exprs:
            if not e.py_value():
                return False
        return True

    def _py_func_binary(self, operands):
        return self._py_func_binary_func(
            operands.exprs[0].py_value(), operands.exprs[1].py_value()
        )

    def _py_func_if(self, operands):
        if operands.exprs[0].py_value():
            return operands.exprs[1].py_value()
        if len(operands.exprs) == 2:
            return False
        return operands.exprs[2].py_value()

    def _py_func_minus(self, operands):
        if len(operands.exprs) == 1:
            return -operands.exprs[0].py_value()
        return operands.exprs[0].py_value() - operands.exprs[1].py_value()

    def _py_func_multi(self, operands):
        def _iter():
            for e in operands.exprs:
                if "cells" in e:
                    for c in e.cells:
                        yield c.expr, c.expr.py_value()
                else:
                    yield e, e.py_value()

        rv = self._py_func_multi_init
        for e, v in _iter():
            if not isinstance(v, decimal.Decimal):
                self._error(
                    "not decimal operand type={} value={} operand={}", type(v), v, e
                )
            rv = v if rv is None else self._py_func_multi_func(rv, v)
        return rv

    def _py_func_not(self, operands):
        return not operands.exprs[0].py_value()

    def _py_func_or(self, operands):
        for e in operands.exprs:
            if e.py_value():
                return True
        return False

    def _xl_func_default(self, operands):
        return (
            f"{self._xl_func_default_name}("
            + ",".join((o.xl_formula() for o in operands.exprs))
            + ")"
        )

    def _xl_func_infix(self, operands):
        return f"{self._xl_infix(operands.exprs[0])}{self.token}{self._xl_infix(operands.exprs[1])}"

    def _xl_func_minus(self, operands):
        if len(operands.exprs) == 1:
            return f"{self.token}{self._xl_infix(operands.exprs[0])}"
        return self._xl_func_infix(operands)

    def _xl_infix(self, expr):
        rv = expr.xl_formula()
        return f"({rv})" if expr.pkunchecked_nested_get("op_spec._infix") else rv


class _Row(_Base):
    def __init__(self, cfg):
        """Creates a row"""
        super().__init__(cfg)
        self.cells = PKDict()

    def add_cell(self, col, content_or_cell, **kwargs):
        """Adds a cell to `col` in `self`

        Args:
            col (str): name of column to add
            content_or_cell (object): See `_Base.cell` for arguments
        """
        if col in self.cells:
            self._error(
                "cell={} already exists in cells={}", col, sorted(self.cells.keys())
            )
        self.cells[col] = (
            self.cell(content_or_cell, **kwargs).pkupdate(col=col)._relations(self)
        )
        return self

    def add_cells(self, *args, **kwargs):
        """Adds `values` to the row/header/footer

        For the footer, the first col must match first header

        Args:
            values (dict or list, tuple): ordered col=label/cell, where col is a keyword name or list of cells
            kwargs (dict): ordered col=label/cell
        Returns:
            self: row/header/footer
        """
        if len(args) == 0:
            # Allow empty case for call to row(), header(), etc.
            c = PKDict() if len(kwargs) == 0 else kwargs
        elif len(args) > 1:
            raise self._error("too many (>1) args={}", args)
        else:
            if isinstance(args[0], dict):
                c = PKDict(args[0])
            elif isinstance(args[0], (tuple, list)):
                # Really only useful for headers case
                c = PKDict(zip(args[0], args[0]))
            else:
                self._error("invalid cell={}", args[0])
            c.update(kwargs)
        for k, v in c.items():
            self.add_cell(k, v)
        return self

    def _children(self):
        return self.cells.values()

    def _compile_pass1(self):
        for k, v in self.cells.items():
            if k not in self.parent.cols:
                self._error("column={} does not exist; cell={}", k, v)
        s = set()
        r = self.row_num
        for i, n in enumerate(self.parent.cols, _COL_NUM_1):
            if n not in self.cells:
                self._error("column={} not found", n)
            self.cells[n].pkupdate(
                col_num=i,
                row_num=r,
                sort_index=i * _ROW_MODULUS + r,
                xl_id=f"{_XL_COLS[i]}{r}",
            )._compile_pass1()

    def _save(self):
        for c in self._children():
            c._save()


class _Footer(_Row):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.defaults.pksetdefault(border="top")


class _Header(_Row):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.defaults.pksetdefault(font="bold")


class _Sheet(_Base):
    def __init__(self, cfg):
        """Appends Sheet to Workbook

        Args:
            cfg (dict): configuration, e.g. title, defaults
        """
        super().__init__(cfg)
        self.tables = []

    def blank_table(self):
        """Create a table with one row, which is blank"""
        self.table(title=f"blank_table-{len(self.tables)}").row()

    def table(self, **kwargs):
        """Appends table to sheets

        Args:
            title (str): debug label for the table
            defaults (PKDict): default values, e.g. round_digits
        """
        return self._child(self.tables, _Table, kwargs)

    def width(self, col, width):
        if col not in self._col_widths or self._col_widths[col] <= width:
            self._col_widths[col] = width

    def _children(self):
        return self.tables

    def _compile_pass1(self):
        r = _ROW_NUM_1
        for t in self._children():
            r = t._compile_pass1(r)

    def _save(self):
        self.xl = self.workbook.xl.add_worksheet(self.title)
        self.text_cells = []
        self._col_widths = PKDict()
        for c in self._children():
            c._save()
        for c, w in self._col_widths.items():
            self.xl.set_column(f"{c}:{c}", width=w)
        if self.text_cells:
            # sqref is represented as list of cells separated by spaces
            self.xl.ignore_errors({"number_stored_as_text": " ".join(self.text_cells)})
        self.xl = None
        self.text_cells = None
        self._col_widths = None


class _Table(_Base):
    def __init__(self, cfg):
        """Append a table to Sheet

        The first row (be it header, row, or footer) defines the
        column names and the maximum width of the table.

        The column names must match this first row. A cell content
        may be None, which is an empty row.

        Args:
            cfg (dict): configuration, e.g. defaults
        """
        super().__init__(cfg)
        self.headers = []
        self.rows = []
        self.footers = []

    def footer(self, *args, **kwargs):
        """Append a footer

        The first footer will be separated from the table with top border.

        Args:
            cells (dict, kwargs, list): ordered col=cell; coll must match first header
        """
        return self._child(self.footers, _Footer, args, kwargs)

    def header(self, *args, **kwargs):
        """Append a header

        The first header defines the column names and the width of the table.

        Args:
            cells (dict, kwargs, list): ordered col=label, where col is a keyword name
        """
        return self._child(self.headers, _Header, args, kwargs)

    def row(self, *args, **kwargs):
        """Append a row

        Args:
            cells (dict, kwargs, list): ordered col=label, where col is a keyword name
        """
        return self._child(self.rows, _Row, args, kwargs)

    def _child(self, children, child, args, kwargs):
        return super()._child(children, child, PKDict()).add_cells(*args, **kwargs)

    def _children(self):
        return self.headers + self.rows + self.footers

    def _compile_pass1(self, row_num):
        self.first_row_num = row_num
        for r in self._children():
            if "cols" not in self:
                self.cols = tuple(r.cells.keys())
                self.col_set = frozenset(self.cols)
            r.row_num = row_num
            r._compile_pass1()
            row_num += 1
        return row_num

    def _save(self):
        for c in self._children():
            c._save()


def _init():

    global _XL_COLS, _DIGITS_TO_PLACES
    if _XL_COLS:
        return
    # really, you can 16384 columns, but we only support 702 (26 + 26*26)
    x = [chr(ord("A") + i) for i in range(26)]
    v = ([None] * _ROW_NUM_1) + x.copy()
    for c in x:
        v.extend([c + d for d in x])
    _XL_COLS = tuple(v)
    x = ("1",) + tuple(
        (("." + ("0" * i) + "1") for i in range(16)),
    )
    _DIGITS_TO_PLACES = tuple((decimal.Decimal(d) for d in x))
    _OpSpec.init()


def _rnd(v, digits):
    return v.quantize(
        _DIGITS_TO_PLACES[digits],
        rounding=decimal.ROUND_HALF_UP,
    )


_init()



# -*- coding: utf-8 -*-
"""test for :mod:`pykern.pkcompat`

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import locale
import os


def setup_module():
    """Set locale so can test expected outputs.

    TODO(robnagler) should test multiple locales.
    """
    # Setting locale as a tuple doesn't work. Not clear this is cross-platform
    os.environ["LANG"] = "en_US.UTF-8"
    locale.setlocale(locale.LC_ALL)


def test_from_bytes():
    from pykern import pkcompat
    from pykern.pkunit import pkeq

    b = pkcompat.to_bytes("你好")
    s = pkcompat.from_bytes(b)
    pkeq(s, "你好")
    pkeq(b, b"\xe4\xbd\xa0\xe5\xa5\xbd")
    pkeq(False, b == s)


def test_locale_str_1():
    """Verify proper conversions"""
    from pykern import pkcompat

    s = pkcompat.locale_str(b"\xc2\xb0")
    assert isinstance(s, str), "When locale_str is converted, it should return str"
    assert "°" == s, "Conversion should be same as literal unicode value"
    before = str(123)
    assert before == pkcompat.locale_str(
        before
    ), "When string is already unicode, conversion yields same string"
    before = str(None)
    assert before == pkcompat.locale_str(
        before
    ), "When string is already unicode, conversion yields same string"


def test_locale_str_2():
    """Invalid utf8"""
    from pykern import pkcompat, pkunit

    with pkunit.pkexcept(UnicodeDecodeError):
        # TODO(robngler) set the locale?
        pkcompat.locale_str(b"\x80")


def test_unicode_unescape():
    from pykern import pkcompat

    assert "\n" == pkcompat.unicode_unescape(r"\n")


def test_zip_strict():
    from pykern import pkcompat, pkunit

    for e, c in (
        ((), ()),
        ((), (range(0), range(0))),
        (((0,),), (range(1),)),
        (((0, 0),), (range(1), range(1))),
        (((0, 0, 0),), (range(1), range(1), range(1))),
    ):
        pkunit.pkeq(e, tuple(pkcompat.zip_strict(*c)))

    for c in (
        (range(0), range(1)),
        (range(0), range(0), range(1)),
        (range(1), range(0)),
        (range(1), range(2)),
        (range(2), range(3)),
    ):
        with pkunit.pkexcept(ValueError):
            list(pkcompat.zip_strict(*c))



"""Initialize Python project directories

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkcli
from pykern import pkio
from pykern import pkjinja
from pykern import pkresource
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdp
import copy
import datetime
import os
import py.path
import re
import subprocess

#: Default values
DEFAULTS = {
    "year": datetime.datetime.now().year,
    "license": "apache2",
    "copyright_license_rst": """:copyright: Copyright (c) {year} {author}.  All Rights Reserved.
:license: {license}""",
}


#: Licenses
LICENSES = {
    "agpl3": (
        "https://www.gnu.org/licenses/agpl-3.0.txt",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ),
    "apache2": (
        "https://www.apache.org/licenses/LICENSE-2.0.html",
        "License :: OSI Approved :: Apache Software License",
    ),
    "gpl2": (
        "https://www.gnu.org/licenses/gpl-2.0.txt",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ),
    "gpl3": (
        "https://www.gnu.org/licenses/gpl-3.0.txt",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ),
    "lgpl2": (
        "https://www.gnu.org/licenses/lgpl-2.0.txt",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
    ),
    "lgpl3": (
        "https://www.gnu.org/licenses/lgpl-3.0.txt",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ),
    "mit": (
        "https://opensource.org/licenses/MIT",
        "License :: OSI Approved :: MIT License",
    ),
    "proprietary": (
        "PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.",
        "License :: Other/Proprietary License",
    ),
}

_UPGRADE_PKSETUP_KEYS = frozenset(
    (
        "author",
        "author_email",
        "classifiers",
        "description",
        "license",
        "name",
        "url",
    )
)


_IGNORE_CLASSIFIERS = frozenset(
    [
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python",
        "Topic :: Utilities",
    ]
    + [c[1] for c in LICENSES.values()],
)


_RS_GIT_URL = "https://git.radiasoft.org/"


def init_rs_tree(description):
    """Initialize defaults for RadiaSoft project tree and call `init_tree`

    `description` is passed verbatim. Other args passed to `init_tree`:

    name
       Base name of the current directory
    author
       "RadiaSoft LLC"
    author_email
       "pip@radiasoft.net"
    license
       "apache2"
    url
       "https://git.radiasoft.org/{name}"

    Args:
        description (str): one-line summary of project
    """
    name = py.path.local().basename
    init_tree(
        name,
        "RadiaSoft LLC",
        "pip@radiasoft.net",
        description,
        "apache2",
        _RS_GIT_URL + name,
    )


def init_tree(name, author, author_email, description, license, url, overwrite=False):
    """Setup a project tree with: docs, tests, etc., and checkin to git.

    Creates: pyproject.toml, index.rst, project dir, <name>_console.py, etc.
    Overwrites files if they exist without checking.

    Args:
        name (str): short name of the project, e.g. ``pykern``.
        author (str): copyright holder, e.g. ``RadiaSoft LLC``
        author_email (str): how to reach author, e.g. ``pip@pykern.org``
        description (str): one-line summary of project
        license (str): name of license, e.g. apache2, mit, and proprietary.
        url (str): website for project, e.g. https://pykern.org
        overwrite (bool): overwrite existing files [False]
    """
    assert os.path.isdir(".git"), "Must be run from the root directory of the repo"
    assert overwrite or not os.path.isdir(
        name
    ), "{}: already exists, only works on fresh repos".format(name)
    assert (
        name == py.path.local().basename
    ), "{}: name must be the name of the current directory".format(name)
    license = license.lower()
    base = pkresource.filename("projex")
    values = copy.deepcopy(DEFAULTS)
    values.update(
        name=name,
        author=author,
        description=description,
        author_email=author_email,
        url=url,
        license=_license(license, 0),
        classifier_license=_license(license, 1),
    )
    values["copyright_license_rst"] = values["copyright_license_rst"].format(**values)
    suffix_re = r"\.jinja$"
    for src in pkio.walk_tree(base, file_re=suffix_re):
        dst = py.path.local(src).relto(str(base))
        dst = dst.replace("projex", name).replace("dot-", ".")
        dst = re.sub(suffix_re, "", dst)
        pkio.mkdir_parent_only(dst)
        _render(src, values, output=dst)
    src = py.path.local(pkresource.filename("projex-licenses"))
    src = src.join(license + ".jinja")
    _render(src, values, output="LICENSE")


def upgrade_tree():
    """Upgrade from setup.py to pyproject.toml

    Replaces all generated files.
    """

    def _classifiers(old_list):
        rv = list(filter(lambda c: c not in _IGNORE_CLASSIFIERS, old_list))
        if rv:
            return ["classifiers in pyproject.toml trimmed, maybe add:"] + rv
        return []

    def _delete_unknown_keys(kwargs):
        x = set(kwargs.keys()) - _UPGRADE_PKSETUP_KEYS
        if not x:
            return []
        return ["Update these keys in pyproject.toml:"] + [
            f"    {k}={kwargs.pkdel(k)}" for k in sorted(x)
        ]

    def _dependencies(install_requires):
        o = pkio.read_text("pyproject.toml")
        n = o.replace(
            '    "pykern"', ",\n".join([f'    "{k}"' for k in install_requires])
        )
        pkio.write_text("pyproject.toml", n)

    def _license_key(url):
        url = url.replace("http:", "https:")
        for k, v in LICENSES.items():
            if v[0] == url:
                return k
        pkcli.command_error("unknown license url={} in setup.py", url)

    def _readme(old_txt, kwargs):
        if old_txt.startswith(
            f"""### {kwargs.name}

{kwargs.description}

Learn more at http"""
        ):
            return []
        pkio.write_text("README.md", old_txt)
        rv = [
            "README.md left intact; update documentation url to:",
            f"    https://{kwargs.name}.readthedocs.io",
        ]
        if kwargs.url.startswith(_RS_GIT_URL):
            rv += [
                "Update project url to:",
                f"    {kwargs.url}",
            ]
        return rv

    def _url(old_url, name):
        if old_url == f"https://github.com/radiasoft/{name}":
            return _RS_GIT_URL + name
        return old_url

    rv = []
    s = _pksetup_kwargs()
    o = PKDict({k: s.pkdel(k) for k in ("classifiers", "install_requires")})
    o.readme = pkio.read_text("README.md")
    rv += _classifiers(o.classifiers)
    rv += _delete_unknown_keys(s)
    s.license = _license_key(s.license)
    s.url = _url(s.url, s.name)
    s.overwrite = True
    init_tree(**s)
    _dependencies(o.install_requires)
    rv += _readme(o.readme, s)
    pkio.unchecked_remove("setup.py", f"{s.name}/base_pkconfig.py")
    return "\n".join(rv + ["Run git diff and then git commit changes"])


def _license(name, which):
    """Returns matching license or command_error"""
    try:
        return LICENSES[name][which]
    except KeyError:
        pkcli.command_error(
            "{}: unknown license name. Valid licenses: {}",
            name,
            " ".join(sorted(LICENSES.values())),
        )


def _pksetup_kwargs():
    """Extra kwargs passed to to setup in setup.py for the old project

    Monkey patches `pksetup.setup` to capture its kwargs.

    Returns:
        PKDict: kwargs passed to setup

    """
    from pykern import pksetup, pkrunpy

    rv = None

    def _save(**kwargs):
        nonlocal rv
        rv = PKDict(kwargs)

    p = getattr(pksetup, "setup")
    try:
        setattr(pksetup, "setup", _save)
        pkrunpy.run_path_as_module("setup.py")
        return rv
    finally:
        setattr(pksetup, "setup", p)


def _render(*args, **kwargs):
    """Renders the template and adds to git"""
    pkjinja.render_file(*args, **kwargs)
    subprocess.check_call(["git", "add", kwargs["output"]])



def force_arg_per_line(
    this,
    be,
    a,
    very,
    poorly,
    formatted,
    function,
    that,
    noone,
    should,
    ever,
    use,
    it,
    has,
    way,
    too,
    many,
    arguments,
):
    foo("force_arg_per_line")



# -*- coding: utf-8 -*-
"""PyTest for :mod:`pykern.pkunit`

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest
import shutil


def test_assert_object_with_json():
    from pykern import pkunit

    pkunit.empty_work_dir()
    pkunit.assert_object_with_json("assert1", {"a": 1})
    with pytest.raises(AssertionError):
        pkunit.assert_object_with_json("assert1", {"b": 1})


def test_pkfail_output():
    from pykern.pkunit import pkexcept, pkne, pkeq
    import re

    k = "example KeyError message"
    with pkexcept(
        re.compile(k + r".*?" + "expecting ValueError but will get KeyError", re.DOTALL)
    ):
        with pkexcept(ValueError, "expecting ValueError but will get {}", "KeyError"):
            raise KeyError(k)

    with pkexcept("expect=x != actual=y"):
        pkeq("x", "y")
    with pkexcept("expect=x != actual=y args: arg1 arg2 kwarg=z"):
        pkeq("x", "y", "args: {} {} kwarg={kwarg}", "arg1", "arg2", kwarg="z")
    with pkexcept("expect=x == actual=x args: arg1 arg2 kwarg=z"):
        pkne("x", "x", "args: {} {} kwarg={kwarg}", "arg1", "arg2", kwarg="z")


def test_data_dir():
    from pykern import pkio
    from pykern import pkunit

    expect = _expect("pkunit_data")
    d = pkunit.data_dir()
    assert isinstance(
        d, type(pkio.py_path())
    ), "Verify type of data_dir is same as returned by py.path.local"
    assert d == expect, "Verify data_dir has correct return value"


def test_data_yaml():
    from pykern import pkunit

    y = pkunit.data_yaml("t1")
    assert "v1" == y["k1"], "YAML is read from file in data_dir"


def test_empty_work_dir():
    from pykern import pkunit
    from pykern import pkio
    import os

    expect = _expect("pkunit_work")
    if os.path.exists(str(expect)):
        expect.remove(rec=1)
    assert not os.path.exists(str(expect)), "Ensure directory was removed"
    d = pkunit.empty_work_dir()
    assert isinstance(
        d, type(pkio.py_path())
    ), "Verify type of empty_work_dir is same as returned by py.path.local"
    assert expect == d, "Verify empty_work_dir has correct return value"
    assert os.path.exists(str(d)), "Ensure directory was created"


def test_file_eq():
    from pykern import pkio
    from pykern import pkunit
    import array

    a = array.ArrayType("d", [1])
    pkunit.file_eq("file_eq1.json", actual=a)
    with pkunit.pkexcept(TypeError):
        pkunit.file_eq("file_eq2.txt", actual=dict())
    d = pkunit.empty_work_dir()
    pkio.write_text(d.join("file_eq3.txt"), "something")
    with pkunit.pkexcept("both exist"):
        pkunit.file_eq("file_eq3.txt", actual="something else")


@pytest.mark.skipif(shutil.which("ndiff") is None, reason="ndiff not available")
def test_file_eq_ndiff():
    from pykern import pkunit

    d = pkunit.data_dir()

    pkunit.file_eq(
        expect_path=d.join("expect_default_conformance.ndiff"),
        actual_path=d.join("actual_default_conformance.ndiff"),
    )
    with pkunit.pkexcept("diffs detected:"):
        pkunit.file_eq(
            expect_path=d.join("expect_default_conformance_1.ndiff"),
            actual_path=d.join("actual_default_conformance_1.ndiff"),
        )
    pkunit.file_eq(
        expect_path=d.join("expect_abs_ok.ndiff"),
        actual_path=d.join("actual_abs_ok.ndiff"),
        ndiff_epsilon_is_abs=True,
        ndiff_epsilon=1e-20,
    )
    with pkunit.pkexcept("diffs detected:"):
        pkunit.file_eq(
            expect_path=d.join("expect_abs_ok.ndiff"),
            actual_path=d.join("actual_abs_ok.ndiff"),
            ndiff_epsilon=1e-20,
        )
    with pkunit.pkexcept("diffs detected:"):
        pkunit.file_eq(
            expect_path=d.join("expect_rel_ok.ndiff"),
            actual_path=d.join("actual_rel_ok.ndiff"),
            ndiff_epsilon_is_abs=True,
        )
    pkunit.file_eq(
        expect_path=d.join("expect_rel_ok.ndiff"),
        actual_path=d.join("actual_rel_ok.ndiff"),
    )
    pkunit.file_eq(
        expect_path=d.join("expect_rel_ok_2.ndiff"),
        actual_path=d.join("actual_rel_ok_2.ndiff"),
        ndiff_epsilon=1e-20,
        ndiff_epsilon_is_abs=False,
    )


def test_file_eq_is_bytes():
    from pykern import pkio
    from pykern import pkunit

    with pkio.save_chdir(pkunit.data_dir()) as d:
        pkunit.file_eq("in.bin", actual_path=d.join("out.bin"), is_bytes=True)
        with pkunit.pkexcept("differ: byte 88"):
            pkunit.file_eq("in.bin", actual_path=d.join("different.bin"), is_bytes=True)
        with pkunit.pkexcept(UnicodeDecodeError):
            pkunit.file_eq("in.bin", actual_path=d.join("out.bin"))


def test_import_module_from_data_dir(monkeypatch):
    from pykern import pkunit

    real_data_dir = pkunit.data_dir()
    fake_data_dir = None

    def mock_data_dir():
        return fake_data_dir

    monkeypatch.setattr(pkunit, "data_dir", mock_data_dir)
    fake_data_dir = str(real_data_dir.join("import1"))
    assert (
        "imp1" == pkunit.import_module_from_data_dir("p1").v
    ), 'import1/p1 should be from "imp1"'
    fake_data_dir = str(real_data_dir.join("import2"))
    assert (
        "imp2" == pkunit.import_module_from_data_dir("p1").v
    ), 'import2/p1 should be from "imp2"'


def test_is_test_run():
    from pykern import pkunit

    pkunit.pkeq(True, pkunit.is_test_run())


def test_pkexcept():
    import re, inspect
    from pykern.pkunit import pkexcept, pkfail

    def _exc_args(pattern, exc):
        if not re.search(pattern, str(exc.args)):
            pkfail("'{}' not found in e.args={}", pattern, exc)

    with pkexcept(KeyError, "should see a KeyError"):
        {}["not found"]
    with pkexcept("KeyError.*xyzzy"):
        {}["xyzzy"]
    try:
        l = inspect.currentframe().f_lineno + 2
        with pkexcept(KeyError, "xyzzy"):
            pass
    except AssertionError as e:
        _exc_args("xyzzy", e)
        _exc_args(f"pkunit_test.py:(?:{l}|{l-1}):test_pkexcept", e)
    except Exception as e:
        pkfail("{}: got exception, but not AssertionError", e)
    else:
        pkfail("did not raise AssertionError")
    try:
        with pkexcept(KeyError):
            raise NameError("whatever")
    except AssertionError as e:
        _exc_args(r"exception was raised.*but expected.*KeyError", e)
    except Exception as e:
        pkfail("{}: got exception, but not AssertionError", e)
    else:
        pkfail("did not raise AssertionError")
    try:
        l = inspect.currentframe().f_lineno + 2
        with pkexcept("any pattern"):
            pass
    except AssertionError as e:
        _exc_args(f"pkunit_test.py:(?:{l}|{l-1}):test_pkexcept", e)
        _exc_args("was not raised", e)
    except Exception as e:
        pkfail("{}: got exception, but not AssertionError", e)
    else:
        pkfail("did not raise AssertionError")


def test_pkok():
    from pykern.pkunit import pkok
    import inspect

    assert 1 == pkok(
        1, "should not see this"
    ), "Result of a successful ok is the condition value"
    lineno = inspect.currentframe().f_lineno + 2
    try:
        pkok(0, "xyzzy {} {k1}", "333", k1="abc")
    except AssertionError as e:
        # May not match exactly, because depends on start directory
        assert "pkunit_test.py:{}:test_pkok xyzzy 333 abc".format(lineno) in str(e.args)


def test_pkre_convert():
    from pykern.pkunit import pkre

    r = r"A"
    for s in ("A", b"A", r"A", "A"):
        pkre(r, s)


def test_xlsx_to_csv_conversion():
    from pykern.pkunit import file_eq, data_dir, empty_work_dir

    file_eq(
        expect_path="example.csv",
        actual_path=data_dir()
        .join("example.xlsx")
        .copy(empty_work_dir().join("example.xlsx")),
    )


def test_ignore_lines():
    from pykern import pkunit

    d = pkunit.data_dir()
    pkunit.file_eq(
        expect_path=d.join("ignore_lines.in"),
        actual_path=d.join("ignore_lines.out"),
        ignore_lines=[r"[0-9]\+xyz"],
    )
    with pkunit.pkexcept("diff command failed"):
        pkunit.file_eq(
            expect_path=d.join("ignore_lines.in"),
            actual_path=d.join("ignore_lines.out"),
            ignore_lines=[r"[invalid regex"],
        )


def _expect(base):
    from pykern import pkio

    d = pkio.py_path(__file__).dirname
    return pkio.py_path(d).join(base).realpath()



# -*- coding: utf-8 -*-
"""restarts infinitely

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_always():
    from pykern import pkunit

    pkunit.restart_or_fail("restart infinitely")



"""test projex upgrade

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_upgrade():
    import subprocess
    from pykern import pkunit, pkio
    from pykern.pkcli import projex

    for d in pkunit.case_dirs():
        subprocess.check_call(["git", "init", "."])
        pkio.write_text("projex.out", projex.upgrade_tree())
        # base_pkconfig setup.py



# -*- coding: utf-8 -*-
u"""test pykern.mpi

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pykern.mpi


def op():
    from mpi4py import MPI
    import time
    import sys

    x = sys.argv[1]
    print(x)
    if 'normal' in x:
        return
    if 'exit-1' == x:
        raise SystemExit(1)
    if 'divide-zero' == x:
        i = 1 / 0
    if 'exit-13-rank-0' == x:
        if MPI.COMM_WORLD and MPI.COMM_WORLD.Get_rank() == 0:
            raise SystemExit(13)
        time.sleep(1)
        return
    if 'divide-zero-rank-2' == x:
        if MPI.COMM_WORLD and MPI.COMM_WORLD.Get_rank() == 2:
            time.sleep(.2)
            i = 1 / 0
        time.sleep(1)
    else:
        raise ValueError('{}: invalid argv'.format(sys.argv))


pykern.mpi.checked_call(op)



# -*- coding: utf-8 -*-
"""test removal of script dir from sys.path

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_sys_path():
    from pykern import pkcli, pkunit, pkdebug
    import sys, os.path

    # py.test adds pwd to sys.path so we have to remove for this test to work
    sys.path.pop(0)
    d = os.path.dirname(os.path.realpath(sys.argv[0]))
    pkunit.pkeq(d, sys.path[0])
    try:
        pkcli.main("pykern")
    except Exception:
        pass
    pkunit.pkne(d, sys.path[0])



"""test github

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
import pytest
import os
import re

pytestmark = pytest.mark.skipif(
    (
        bool(os.environ.get("PYKERN_TEST_CI"))
        or not bool(os.environ.get("PYKERN_PKCLI_GITHUB_PASSWORD"))
    ),
    reason="CI pipelines use shared IPs so gets rate limited too easily",
)


def test_repo_no_wiki():
    from pykern import pkconfig
    from pykern.pkdebug import pkdlog, pkdp, pkdc, pkdexc

    pkconfig.reset_state_for_testing(
        {
            "PYKERN_PKCLI_GITHUB_TEST_MODE": "1",
            "PYKERN_PKCLI_GITHUB_API_PAUSE_SECONDS": "0",
        }
    )
    import os
    from pykern.pkcli import github
    from pykern import pkunit
    from pykern import pkio
    import subprocess

    with pkunit.save_chdir_work():
        github.backup("radiasoft")
        for n, e in PKDict(
            {
                "radiasoft-test-pykern-github": 1,
                "radiasoft-test-pykern-github-no-wiki": 0,
                "biviosoftware-test-pykern-github-no-wiki": 0,
            }
        ).items():
            a = len(pkio.sorted_glob(f"**/{n}.wiki.git/config"))
            pkunit.pkeq(e, a, "repo={} expected={} != actual={}", n, e, a)


def test_backup():
    from pykern import pkconfig

    pkconfig.reset_state_for_testing(
        {
            "PYKERN_PKCLI_GITHUB_TEST_MODE": "1",
            "PYKERN_PKCLI_GITHUB_API_PAUSE_SECONDS": "0",
        }
    )
    from pykern.pkcli import github
    from pykern import pkunit
    from pykern import pkio

    with pkunit.save_chdir_work():
        github.backup("radiasoft")
        github.backup("radiasoft")
        x = pkio.sorted_glob("*/radiasoft-test-pykern-github.git/config")
        pkunit.pkeq(2, len(x))
        pkunit.pkeq(2, x[0].stat().nlink)
        x = [
            os.popen(f"tar tJf {y}").read()
            for y in pkio.sorted_glob("*/radiasoft-test-pykern-github.issues.txz")
        ]
        pkunit.pkeq(2, len(x))
        pkunit.pkeq(x[0], x[1])


def test_issue_start():
    from pykern import pkconfig

    pkconfig.reset_state_for_testing(
        {
            "PYKERN_PKCLI_GITHUB_TEST_MODE": "1",
            "PYKERN_PKCLI_GITHUB_API_PAUSE_SECONDS": "0",
        }
    )
    from pykern.pkcli import github
    from pykern import pkunit

    r = _close_issues()
    github.issue_pending_alpha(_test_repo())
    with pkunit.pkexcept("prior release"):
        github.issue_start_alpha(_test_repo())

    def _commit(repo, full_name, issues):
        issues.append(_create_commit(repo, full_name=full_name))
        a = github.issue_update_alpha_pending(_test_repo())
        pkunit.pkre(issues[-1].title, a)
        pkunit.pkre(issues[-1].link, a)

    issues = []
    _commit(r, False, issues)
    _commit(r, True, issues)

    m = None
    for x in (
        github.issue_start_alpha,
        github.issue_start_beta,
        github.issue_start_prod,
    ):
        a = x(_test_repo())
        for i in issues:
            pkunit.pkre(i.link, a)
            pkunit.pkre(i.title, a)

        if m:
            pkunit.pkre(m.group(1), a)
        m = re.search(r"(?:Started|Created) (#\d+)", a)
        assert m
        r.issue(m.group(1)[1:]).close()
    github.issue_pending_alpha(_test_repo())
    a = github.issue_update_alpha_pending(_test_repo())
    pkunit.pkeq("", a)


def test_labels():
    from pykern import pkconfig

    pkconfig.reset_state_for_testing(
        {
            "PYKERN_PKCLI_GITHUB_TEST_MODE": "1",
            "PYKERN_PKCLI_GITHUB_API_PAUSE_SECONDS": "0",
        }
    )
    from pykern.pkcli import github
    from pykern import pkunit

    github.labels(_test_repo())
    github.labels(_test_repo(), clear=True)


def test_list_repos():
    from pykern.pkcli import github
    from pykern import pkunit

    r = github.list_repos("radiasoft")
    pkunit.pkok("pykern" in r, "pykern missing in actual={}", r)


def _close_issues():
    from pykern.pkcli import github

    r = github.GitHub().repo(_test_repo())
    for i in r.issues(state="open"):
        if re.search(" release ", i.title, flags=re.IGNORECASE):
            i.close()
    return r


def _create_commit(repo, full_name=False):
    from pykern import pkcompat
    from pykern import pkunit, util

    t = util.random_base62()
    i = repo.create_issue(title=t + " for github_test", body="n/a")
    m = repo.readme()
    b = pkcompat.from_bytes(m.decoded)
    x = re.sub(r"(?<=github_test)=([^\n]+)", t, b)
    if b == x:
        b += f"github_test={t}\n"
    r = f'{repo if full_name else ""}#{i.number}'
    m.update(f"fix {r} for github_test", pkcompat.to_bytes(b))
    return PKDict(link=f"{repo}#{i.number}", title=t)


def _test_repo():
    from pykern.pkcli import github

    t = github._TEST_REPOS[0]
    return "/".join((t.org, t.name))



"""run test files in separate processes

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkconfig
from pykern import pkconst
from pykern import pkio
from pykern import pkunit
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdp
import os
import pykern.pkcli
import re
import signal
import subprocess
import sys
import time


SUITE_D = "tests"
_COROUTINE_NEVER_AWAITED = re.compile(
    "(.+ coroutine \S+ was never awaited.*)", flags=re.MULTILINE
)
_TEST_SKIPPED = re.compile(r"^.+\s+SKIPPED\s+\(.+\)$", flags=re.MULTILINE)
_TEST_PY = re.compile(r"_test\.py$")

_FAIL_MSG = "FAIL"

_PASS_MSG = "pass"

_SKIPPED_MSG = "skipped"

_MAX_RESTARTS = 3

_RESTART_MSG = "restart"

_START_MSG = "start"

_WAIT_LOOP_SLEEP = 0.1

_cfg = pkconfig.init(
    ignore_warnings=(False, bool, "override pytest's output of all warnings"),
    max_failures=(
        5,
        pkconfig.parse_positive_int,
        "maximum number of test failures before exit",
    ),
    max_procs=(1, pkconfig.parse_positive_int, "maximum number of parallel test runs"),
    restartable=(
        False,
        bool,
        "allow pkunit.restart_or_fail to restart",
    ),
)


def default_command(*args):
    """Run tests one at a time with py.test.

    Searches in ``tests`` sub-directory if not provided a
    list of tests or not in tests/ already

    Arguments are directories or files, which are searched for _test.py
    files.

    An argument which is ``case=<pattern>``, is passed to pytest
    as ``-k <pattern>``.

    ``skip_past=<last_to_ignore>`` causes collection to ignore all
    files up to and including ``<last_to_ignore>``` (may be partial
    match of the test file).

    ``max_procs=2`` starts two cases simultaneously. Default is 1,
    or config value `PYKERN_PKCLI_TEST_MAX_PROCS` if set.

    Writes failures to ``<base>_test.log``

    Args:
        args (str): test dirs, files, options
    Returns:
        str: passed=N if all passed, else raises `pkcli.Error`

    """
    return _Runner(args).result


class _Case:
    def __init__(self, rel_path, runner):
        super().__init__()
        self.runner = runner
        self.rel_path = rel_path
        self.abs_path = pkio.py_path(rel_path)
        self.tries = _MAX_RESTARTS if _cfg.restartable else 1
        self.run()

    def is_done(self, aborting):
        if (e := self.process.poll()) is None:
            return None
        return self._exit(e, aborting)

    def pkdebug_str(self):
        return f"{self.__class__.__name__}({self.rel_path})"

    def run(self):
        self.tries -= 1
        self.restartable = self.tries > 0
        self.process = self._start()

    def _exit(self, returncode, aborting):
        def _msg():
            if 0 == returncode:
                return _PASS_MSG
            if 5 == returncode:
                # 5 means test was skipped
                # see http://doc.pytest.org/en/latest/usage.html#possible-exit-codes
                return _SKIPPED_MSG
            if not aborting and self.restartable and 2 == returncode:
                # 2 means KeyboardInterrupt
                # POSIT: pkunit.restart_or_fail uses this
                return _RESTART_MSG
            return _FAIL_MSG

        def _skipped(ouput):
            return _TEST_SKIPPED.findall(ouput)

        self.runner.signal_cascade.processes.remove(self.process)
        self.process = None
        o = pkio.read_text(self.output_path)
        self.skipped = _skipped(o) if _TEST_SKIPPED.search(o) else []
        if m := re.findall(_COROUTINE_NEVER_AWAITED, o):
            with self.output_path.open(mode="a") as f:
                f.write("".join(f"ERROR: {x}\n" for x in m))
                # never restartable, it's always a defect
                return _FAIL_MSG
        self.exit = _msg()
        return self.exit

    def _start(self):
        def _env():
            res = os.environ.copy()
            res.update(
                {
                    pkunit.TEST_FILE_ENV: str(self.abs_path),
                    pkunit.RESTARTABLE: "1" if self.restartable else "",
                }
            )
            return res

        def _ignore_warnings():
            if not _cfg.ignore_warnings:
                return []
            rv = []
            for w in (
                # https://docs.python.org/3/library/warnings.html#default-warning-filter
                "DeprecationWarning",
                "PendingDeprecationWarning",
                "ImportWarning",
                "ResourceWarning",
            ):
                rv.extend(("-W", f"ignore::{w}"))
            return rv

        def _remove_work_dir():
            w = _TEST_PY.sub(pkunit.WORK_DIR_SUFFIX, self.rel_path)
            if w != self.rel_path:
                pkio.unchecked_remove(w)

        def _process():
            c = (
                ["pytest"]
                + _ignore_warnings()
                + [
                    "--tb=native",
                    "-v",
                    "-s",
                    "-rs",
                    self.rel_path,
                ]
                + self.runner.pytest_flags
            )
            v = _env()
            with self.output_path.open("w") as o, open(os.devnull) as i:
                return subprocess.Popen(
                    c,
                    stdin=i,
                    stdout=o,
                    stderr=subprocess.STDOUT if o else None,
                    env=v,
                    close_fds=True,
                )

        self.output_path = self.abs_path.new(ext=".log")
        _remove_work_dir()
        pkio.unchecked_remove(self.output_path)
        rv = _process()
        self.runner.signal_cascade.processes.add(rv)
        return rv


class _Runner:
    def __init__(self, args):

        def _too_many_failures():
            if len(self.failures) < _cfg.max_failures:
                return False
            self._info(
                None,
                [
                    f"too many failures={len(self.failures)} aborting"
                    + (
                        "; waiting for processes to finish"
                        if self.max_procs > 1 and self.cases
                        else ""
                    ),
                ],
            )
            return True

        self._args(args)
        if not self.rel_paths:
            pykern.pkcli.command_error("no tests found")
        c = 0
        self.failures = []
        self.cases = set()
        with _SignalCascade() as self.signal_cascade:
            for p in self.rel_paths:
                c += 1
                self._run(p)
                if a := _too_many_failures():
                    break
            while self._wait_for_one(aborting=a):
                pass
        self._assert_failures(self.failures, c)
        self.result = f"passed={c}"

    def _args(self, tests):
        def _file(path):
            if self.skip_past:
                if self.skip_past in path:
                    self.skip_past = None
                return
            self.rel_paths.append(path)

        def _find(paths):
            i = re.compile(r"(?:_work|_data)/")
            cwd = pkio.py_path()
            for t in _resolve_test_paths(paths, cwd):
                t = pkio.py_path(t)
                if not t.exists():
                    pykern.pkcli.command_error("test={} does not exist", t)
                if t.check(file=True):
                    _file(str(cwd.bestrelpath(t)))
                    continue
                for p in pkio.walk_tree(t, _TEST_PY):
                    p = str(cwd.bestrelpath(p))
                    if not i.search(p):
                        _file(p)

        def _flag(name, value):
            if len(value) <= 0:
                pykern.pkcli.command_error(f"empty value for option={name}")
            elif name == "case":
                self.pytest_flags.extend(("-k", value))
            elif name == "max_procs":
                try:
                    self.max_procs = int(value)
                except Exception:
                    self.max_procs = -1
                if self.max_procs <= 0:
                    pykern.pkcli.command_error(
                        f"max_procs={value} must be a positive integer"
                    )
            elif name == "skip_past":
                if self.skip_past:
                    pykern.pkcli.command_error(
                        f"skip_past={value} passed twice (skip_past={self.skip_past})"
                    )
                self.skip_past = value
            else:
                pykern.pkcli.command_error(f"unsupported option={name}")

        def _resolve_test_paths(paths, current_dir):
            if not paths:
                p = current_dir
                if p.basename != SUITE_D:
                    p = SUITE_D
                paths = (p,)
            return paths

        p = []
        self.pytest_flags = []
        self.max_procs = _cfg.max_procs
        self.skip_past = None
        for t in tests:
            if "=" in t:
                _flag(*(t.split("=")))
            else:
                p.append(t)
        self.rel_paths = []
        _find(p)

    def _assert_failures(self, failures, count):
        if len(failures) <= 0:
            return
        # Avoid dumping too many test logs
        for c in failures[: _cfg.max_failures]:
            self._info(None, [pkio.read_text(c.output_path)])
        pykern.pkcli.command_error(
            f"FAILED={len(failures)} passed={count - len(failures)}",
        )

    def _exit(self, case, msg):
        if rv := msg != _RESTART_MSG:
            self.cases.remove(case)
            if msg == _FAIL_MSG:
                self.failures.append(case)
        self._info(case, [msg] + case.skipped)
        case.skipped = None
        return rv

    def _info(self, case, lines):
        if case is None:
            if not lines[-1].endswith("\n"):
                # other output on its own line, ensure newline at end
                lines[-1] += "\n"
        else:
            if lines[0] == _FAIL_MSG:
                # add the failure context
                lines[0] += f" {case.output_path}"
            if self.max_procs > 1:
                # line by line when multiprocess
                lines[0] = case.rel_path + " " + lines[0]
            elif lines[0] == _START_MSG:
                # starting a case
                lines[0] = case.rel_path
            else:
                # completing a case
                lines[0] = " " + lines[0]
                lines[-1] += "\n"
        o = "\n".join(lines)
        if self.max_procs > 1 and not o.endswith("\n"):
            o += "\n"
        pkconst.builtin_print(o, end="")
        # TODO(robnagler) is this necessary?
        sys.stdout.flush()

    def _run(self, rel_path):
        c = _Case(rel_path, self)
        self.cases.add(c)
        self._info(c, [_START_MSG])
        if len(self.cases) >= self.max_procs:
            self._wait_for_one()

    def _wait_for_one(self, aborting=False):
        while self.cases:
            for c in self.cases:
                if (m := c.is_done(aborting)) is None:
                    continue
                if self._exit(c, m):
                    return True
                c.run()
            time.sleep(_WAIT_LOOP_SLEEP)
        return False


class _SignalCascade:
    _SIGNALS = (signal.SIGTERM, signal.SIGINT)
    _IGNORE_HANDLERS = frozenset((None, signal.SIG_IGN, signal.SIG_DFL))

    def __init__(self):
        self.processes = set()

    def __enter__(self):
        self._prev_handlers = PKDict({s: signal.getsignal(s) for s in self._SIGNALS})
        for s in self._prev_handlers.keys():
            signal.signal(s, self._handler)
        return self

    def __exit__(self, type, value, traceback):
        for s, h in self._prev_handlers.items():
            signal.signal(s, h)

    def _handler(self, sig, frame):
        for p in self.processes:
            try:
                p.send_signal(sig)
            except Exception:
                pass
        if (h := self._prev_handlers.get(sig)) not in self._IGNORE_HANDLERS:
            h(sig, frame)



# -*- coding: utf-8 -*-
"""

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest

pytest.importorskip("mpi4py")


def test_checked_call():
    from pykern import pkunit
    from pykern.pkunit import pkeq
    from pykern import pkconst
    import sys
    import subprocess

    with pkunit.save_chdir_work():
        cmd = [sys.executable, str(pkunit.data_dir().join("p1.py"))]
        for i, a in enumerate(
            (
                ("normal", 0),
                ("exit-1", 1),
                ("divide-zero", 1),
                ("normal-rank-all", 0),
                ("divide-zero-rank-2", 86),
                ("exit-13-rank-0", 13),
            )
        ):
            f = "{}.out".format(i)
            with open(f, "w") as o:
                c = cmd + [a[0]]
                pkconst.builtin_print(a[0])
                if "rank" in a[0]:
                    c = ["mpiexec", "-n", "4"] + c
                actual = subprocess.call(
                    c,
                    stdout=o,
                    stderr=subprocess.STDOUT,
                )
            pkeq(a[1], actual, "{}: exit({})\n{}", " ".join(c), actual, open(f).read())



"""sleeps so other tests can run

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_sleep():
    import time

    time.sleep(1)



# -*- coding: utf-8 -*-
"""Front-end command line for :mod:`private`.

See :mod:`pykern.pkcli` for how this module is used.

:copyright: Copyright (c) 2024 James Bond.  All Rights Reserved.
:license: PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.
"""
import pykern.pkcli
import sys


def main():
    return pykern.pkcli.main("private")


if __name__ == "__main__":
    sys.exit(main())



last_cmd = None


def cmd1(arg1):
    """Subject line for cmd1

    Args:
        arg1
    """
    global last_cmd
    last_cmd = cmd1.__name__
    return


def cmd2():
    """Subject line for cmd2

    Args:
        -
    """
    global last_cmd
    last_cmd = cmd2.__name__
    return



# -*- coding: utf-8 -*-
"""pkconfig init

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_init(pkconfig_setup):
    """Validate parse_set"""
    pkconfig = pkconfig_setup(
        cfg=dict(P1_M1_SET3="", P1_M1_SET4="a:b"),
        env=dict(P1_M1_REQ8="99", PYKERN_PKCONFIG_CHANNEL="dev"),
    )
    from p1.m1 import cfg
    from pykern import pkunit

    pkunit.pkeq(set(), cfg.set1)
    pkunit.pkeq(set([1]), cfg.set2)
    pkconfig.reset_state_for_testing()
    pkunit.pkeq(set(), cfg.set3)
    pkunit.pkeq(set(("a", "b")), cfg.set4)
    pkunit.pkeq(True, pkconfig.in_dev_mode())



# -*- coding: utf-8 -*-
"""Logging or regex-controlled print statements

We take the view that there are two types of logging: permanent
(`pkdp`) and dynamically controlled (`pkdc`). In both cases, you want
to know where the print statement originated from. Python's `logging`
module supports logging the origin, but it doesn't support user-
defined dynamic filters in the form of a regular expression.

Permanent print statements (`pkdp`) is what people traditionally think
of as "logging". What we don't care about is "logging levels", because
one person's "critical" is another person's "whatever". It's up to
operations people to figure out what's an error worth paying attention
to, and what's not. That's managed by external rule bases, not inside
some library module.

Python's logging module gives programmers too many options and
adds complexity to filters. If you don't attach a handler to
the loggers, you'll never see some messages. Or, if the handler
has `logging.WARN` set, and you really want to see `logging.INFO`
messages, you won't see them.

There are cases when you really don't want to see output. We call
these dynamically controlled print statements. They are typically used
by developers when debugging a new module, and sometimes these lines
get left in the code for a while so that you can get detailed (noisy)
debugging information on production systems.

Dynamically controlled print statements (`pkdc`) would overwhelm logs
so they are off by default and can be filtered by regular expressions
supplied via configuration (including environment variables) or
programmatically via `init`.

Examples:

    In a module, you would write::

        from pykern.debug import pkdc, pkdexc, pkdp

        pkdp('user entered: {}', val)
        pkdc('user context: name={name}, id={id}', **user_rec)

    The first print statement always appears.
    The second does not unless you specify a "control" via the
    environment variable ``$PYKERN_PKDEBUG_CONTROL``:

        PYKERN_PKDEBUG_CONTROL=my_mod python my_prog.py

    or with the shortname pkdebug::

        pkdebug=my_mod python my_prog.py

    Or, if you want a specific conditional print::

        PYKERN_PKDEBUG_CONTROL=my_mod.py:52:

    You can match any text in the line output with a regular expression, which
    is case insensitive.

    If you omit args and kwargs, `pkdp` returns its argument, of use when you
    want to inspect values within the normal flow of the code, e.g.::

        def my_function():
            ...
            return pkdp(res)

    or

        a = pkdp(my_function_1()) + pkdp(my_function_2())

NOTE: format arguments are converted to strings by this module, and
not by `str.format`. Then they are passed to `str.format`. This means
that you should assume the argument is a string or, better, just a
default type. Arguments are trunctated by `pkdformat`.

If `output` is a string, will open the file to write to. The initial
value of output is ``$PYKERN_PKDEBUG_OUTPUT``.

:copyright: Copyright (c) 2014-2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html

"""
from pykern import pkcompat
from pykern import pkconfig
from pykern import pkconst
from pykern import pkinspect
import datetime
import functools
import inspect
import json
import logging
import numbers
import os
import pprint
import re
import six
import sys
import threading
import traceback


#: Maximum number of exceptions thrown before printing stops
MAX_EXCEPTION_COUNT = 5

#: Was control initialized?
_have_control = False

#: Object which does the writing, initialized every time :func:`init` is called.
_printer = None

#: Get IPython InteractiveShell.write()
# See https://github.com/ipython/ipython/blob/master/IPython/core/interactiveshell.py)
_ipython_write = None
try:
    _ipython_write = get_ipython().write_err
except Exception:
    pass

#: Type of a regular expression
_RE_TYPE = type(re.compile(""))

#: How to parse thread names
_THREAD_ID_RE = re.compile(r"Thread-(\d+)", re.IGNORECASE)

#: name of function objects can implement to control their own string
# representation
PKDEBUG_STR_FUNCTION_NAME = "pkdebug_str"

#: redeacted value marker
REDACTED = "<REDACTED>"

#: keys whose values will be redacted from logs
SECRETS_RE = re.compile(r"(?:secret|otp\b|passw|private)", re.IGNORECASE)

#: truncated string marker
SNIP = "<SNIP>"


def init(**kwargs):
    """May be called to (re)initialize this module.

    `control` is a regular expression, which is used to control the
    output of :func:`pkdc`. Messages from :func:`pkdp` and :func:`pkdc`
    are written to `output`.

    `output` is either an object which implements `write` or a `str`, in which
    case it is opened with :func:`io.open`.

    Args:
        control(str or re.RegexObject): lines matching will be output
        output (str or file): where to write messages [error output]
        redirect_logging (bool): Redirect Python's logging to output [True]
        want_pid_time (bool): display PID and time in messages [False]
    """
    global _printer
    global _have_control
    _printer = _Printer(**kwargs)
    _have_control = _printer.have_control


def pkdc(fmt, *args, **kwargs):
    """Conditional print a message to `output` selectively based on `control`.

    Controlled output that you can leave in your code permanently.

    Args:
        fmt (str): how to :func:`str.format`
        args: what to format
        kwargs: what to format
    """
    # Since calls are left in for product, this check has
    # some value.
    if _have_control:
        _printer._write(fmt, args, kwargs, with_control=True)


def pkdexc(exc_info=None):
    """Return last exception and stack as a string

    Must be called from an ``except``. This function removes
    the last two calls from the stack. It joins the traceback
    so you get a complete stack trace.

    Will catch exceptions during the formatting and returns a
    string in all cases.

    Example::

        try:
            something
        except:
            pkdlog("exception={}", pkdexc())

    Args:
        exc_info (tuple): (type(e), e, traceback) [sys.exc_info]
    Returns:
        str: formatted exception and stack trace
    """
    try:
        e = exc_info or sys.exc_info()
        if hasattr(traceback, "TracebackException"):
            return "".join(
                traceback.format_exception(*e)
                + ["\nException was printed at:\n\n"]
                + traceback.format_exception_only(e[0], e[1])
                + traceback.format_stack()[:-2],
            )

        else:
            return "".join(
                traceback.format_exception_only(e[0], e[1])
                + traceback.format_stack()[:-2]
                + traceback.format_tb(e[2]),
            )

    except Exception as e:
        return "pykern.pkdebug.pkdexc: unable to retrieve exception info"


def pkdformat(fmt, *args, **kwargs):
    """Truncate and redact args & kwargs and then format

    NOTE: converts all elements to strings before calling `str.format`.

    Args:
        fmt (str): how to format
        args (list): what to format
        kwargs (dict): what to format

    Returns:
        str: formatted output
    """
    try:
        args = (_format_arg(a) for a in args)
        kwargs = dict(((k, _format_arg(v)) for k, v in kwargs.items()))
        return fmt.format(*args, **kwargs)
    except Exception:
        _printer.exception_count += 1
        return "invalid format format={} args={} kwargs={} stack={}".format(
            fmt,
            args,
            kwargs,
            pkdexc(),
        )


def pkdlog(fmt_or_arg, *args, **kwargs):
    """Print messages that are intended to be permanent logging.

    See `pkdp` for usage.
    """
    # implementation aliased below
    pass


def pkdp(fmt_or_arg, *args, **kwargs):
    """Print a message to `output` unconditionally, possibly returning its arg

    Use for print statements in your code that you don't intend
    to keep in the system.

    Use `pkdlog` for permanent log messages.

    Args:
        fmt_or_arg (object): how to :func:`str.format`, or object to print
        args: what to format
        kwargs: what to format

    Returns:
        object: Will return fmt_or_arg, if args and kwargs are empty
    """
    if args or kwargs:
        _printer._write(fmt_or_arg, args, kwargs, with_control=False)
    else:
        _printer._write("{}", [fmt_or_arg], {}, with_control=False)
        return fmt_or_arg


# Alias since they are the same
pkdlog = pkdp


def pkdppretty(obj):
    """Print `obj` using `pkdpretty` and return `obj`

    Args:
        obj (object): JSON string or python object

    Returns:
        object: `obj`
    """
    pkdp("{}", pkdpretty(obj))
    return obj


def pkdpretty(obj):
    """Return pretty print the object.

    If `obj` is JSON, parse and print it. If it is a regular python
    data structure, pretty print that. Any exceptions are caught,
    and the return value will be `obj`.

    Args:
        obj (object): JSON string or python object

    Returns:
        str: pretty printed string
    """
    try:
        if isinstance(obj, six.string_types):
            try:
                obj = json.loads(obj)
            except Exception:
                pass
        # try to dump as JSON else dump as Python
        try:
            return (
                json.dumps(
                    obj,
                    sort_keys=True,
                    indent=4,
                    separators=(",", ": "),
                )
                + "\n"
            )
        except Exception as e:
            pass
        if pprint.isreadable(obj):
            return pprint.pformat(obj, indent=4) + "\n"
    except Exception:
        pass
    return obj


class _LoggingHandler(logging.Handler):
    """Handler added to root logger."""

    def emit(self, record):
        """Emit a log record via _printer

        Writes all `logging.INFO` and above like `pkdp`, that is,
        always. Below `logging.INFO` (i.e. DEBUG) is written like
        `pkdc` using the same matching algorithms by converting
        log records appropriately.
        """
        wc = record.levelno < logging.INFO
        _printer._process(
            lambda: pkinspect.Call(record),
            lambda: "{}:{}:{}".format(
                record.levelname, record.name, self.format(record)
            ),
            lambda: (
                record.process,
                datetime.datetime.utcfromtimestamp(record.created),
            ),
            with_control=wc,
        )


class _Printer(object):
    """Internal implementation of :func:`init`. Don't call directly."""

    def __init__(self, **kwargs):
        self.too_many_exceptions = False
        self.exception_count = 0
        # Safe values are set here so we can get over initialization errors
        for k in cfg:
            setattr(self, k, cfg[k])
        self.logging_handler = None
        try:
            self.want_pid_time = self._init_want_pid_time(kwargs)
            self.output = self._init_output(kwargs)
            self.redirect_logging = self._init_redirect_logging(kwargs)
            self.control = self._init_control(kwargs)
            self.have_control = bool(self.control)
        except Exception:
            for k in cfg:
                setattr(self, k, cfg[k])
            self._err("initialization failed, reverting values", pkdexc())
        self._logging_install()

    def _err(self, msg, exc):
        """When a logging error occurs."""
        self.exception_count += 1
        self._out("pykern.pkdebug error: " + msg + "\n" + exc)

    def _init_control(self, kwargs):
        try:
            if "control" in kwargs:
                return _cfg_control(kwargs["control"])
        except Exception:
            self._err("control compile error, using safe value", pkdexc())
        return cfg.control

    def _init_output(self, kwargs):
        try:
            if "output" in kwargs:
                return _cfg_output(kwargs["output"])
        except Exception:
            self._err("output could not be opened, using safe value", pkdexc())
        return cfg.output

    def _init_redirect_logging(self, kwargs):
        return bool(kwargs.get("redirect_logging", cfg.redirect_logging))

    def _init_want_pid_time(self, kwargs):
        return bool(kwargs.get("want_pid_time", cfg.want_pid_time))

    def _logging_install(self):
        """Initialize logging based on redirect_logging"""
        self.logging_handler = None
        self.logging_prev_handlers = None
        self.logging_prev_level = None
        try:
            if _printer:
                _printer._logging_uninstall()
            if not self.redirect_logging:
                return
            # Optimization: Loggers check the level first before creating
            # the LogRecord, just like pkdc checks _have_control so don't
            # want to create LogRecord unnecessarily
            rl = logging.getLogger()
            self.logging_prev_level = rl.level
            self.logging_prev_handlers = []
            while rl.handlers:
                h = rl.handlers[0]
                rl.removeHandler(h)
                self.logging_prev_handlers.append(h)
            level = logging.DEBUG if self.have_control else logging.INFO
            self.logging_handler = _LoggingHandler(level=level)
            rl.addHandler(self.logging_handler)
            rl.setLevel(level)
        except Exception:
            self._err("unable to install logging handler", pkdexc())

    def _logging_uninstall(self):
        """Remove handler from logging stack and set level to previous"""
        try:
            if not self.logging_handler:
                return
            rl = logging.getLogger()
            if not self.logging_prev_level is None:
                rl.setLevel(self.logging_prev_level)
            for h in self.logging_prev_handlers:
                rl.addHandler(h)
            rl.removeHandler(self.logging_handler)
        except Exception:
            pass
        self.logging_handler = None
        self.logging_prev_handlers = None
        self.logging_prev_level = None

    def _out(self, msg):
        """Writes msg to output (or error output if not output)

        If running in IPython, then use ``get_ipython().write_err()``
        so that logging comes out in the cell as an error. Otherwise,
        use stderr.

        If an error occurs, output is reset to None.

        Args:
            output (file): where to write
            msg (str): what to write
        """
        try:
            output = self.output
            if not output:
                if _ipython_write:
                    _ipython_write(msg)
                    return
                output = sys.stderr
            output.write(msg)
            if hasattr(output, "flush"):
                output.flush()
        except Exception as e:
            self.exception_count += 1
            sys.__stderr__.write("output error: " + str(e))
            sys.__stderr__.flush()

    def _pid_time(self, pid, time):
        """Creates pid-time string for output

        Args:
            pid (int): process id
            time (datetime): when did it happen (UTC)

        Returns:
            str: formatted
        """
        if not self.want_pid_time:
            return ""
        try:
            # Force the thread id to a reasonable length so that
            # we don't clutter the logs. It can't be used for anything
            # other than identifying "in the small" log line relationships.
            i = self._thread_id() % 99999
            return "{:%b %d %H:%M:%S} {:5d} {:5d} ".format(time, pid, i)
        except Exception:
            self.exception_count += 1
            self._err("error formatting pid and time", pkdexc())
            return "Xxx 00 00:00:00 00000.0"

    def _prefix(self, call):
        """Format prefix line from location details

        Args:
            caller: state of call
        Returns:
            str: formatted prefix
        """
        return "{} ".format(call)

    def _process(self, call, message, pid_time_values, with_control):
        """Writes formatted message to output with location prefix.

        If not `with_control`, always writes message to
        :attr:`output`. If `with_control` and whole expression matches
        :attr:`control`, writes message, else nothing is output.

        Args:
            call (func): returns filename, line, funcname
            message (func): returns message with prefix as string
            pid_time_values (func): returns pid and time
            with_control (bool): respect :attr:`control`
        """
        if self.too_many_exceptions or with_control and not self.control:
            return
        try:
            msg = self._prefix(call()) + message()
            if not with_control or self.control.search(msg):
                self._out(self._pid_time(*pid_time_values()) + msg.rstrip() + "\n")
        except Exception:
            self._err("unable to process message", pkdexc())
        finally:
            if self.exception_count >= MAX_EXCEPTION_COUNT:
                self.too_many_exceptions = True

    def _thread_id(self):
        """Returns a number to identify the current thread

        Returns:
            int: some number that uniquely identifies the thread
        """
        t = threading.current_thread()
        n = t.name
        if n == "MainThread":
            return 0
        m = _THREAD_ID_RE.search(t.name)
        if m:
            return int(m.group(1))
        return t.ident

    def _write(self, fmt, args, kwargs, with_control=False):
        """Provides formatter for message to _process

        Args:
            fmt_or_record (str or LogRecord): how to format
            args (list): what to format
            kwargs (dict): what to format
            with_control (bool): respect :attr:`control`
        """

        def msg():
            return pkdformat(fmt, *args, **kwargs)

        def pid_time():
            return (os.getpid(), datetime.datetime.utcnow())

        def prefix():
            return pkinspect.Call(
                kwargs.get("pkdebug_frame")
                or inspect.currentframe().f_back.f_back.f_back.f_back,
            )

        self._process(prefix, msg, pid_time, with_control)


@pkconfig.parse_none
def _cfg_control(anything):
    if anything is None:
        return None
    if isinstance(anything, _RE_TYPE):
        return anything
    return re.compile(anything, flags=re.IGNORECASE)


@pkconfig.parse_none
def _cfg_output(anything):
    if anything is None:
        return None
    if hasattr(anything, "write"):
        return anything
    return open(anything, "w")


def _format_arg(obj, depth=0):
    """Redact and truncate `obj`

    Args:
        obj (object): Any argument
        depth (int): nested data structure depth [default: 0]
    Returns:
        object: Redacted and truncated str or obj in exception
    """

    def _dict(obj, depth):
        return _iterate(
            sorted(obj),
            lambda k: _format_arg(k, depth)
            + ": "
            + (_redacted(k) or _format_arg(obj[k], depth)),
            obj,
        )

    def _iterate(sequence, format_fn, obj):
        r = ""
        for i, v in enumerate(sequence):
            if i >= cfg.max_elements:
                break
            r += format_fn(v) + ", "
        return _snip(r, len(obj))

    def _object(obj, depth):
        depth += 1
        c = str(type(obj)()) if isinstance(obj, (list, tuple)) else "{}"
        if depth > cfg.max_depth:
            return c[0] + SNIP + c[1]
        m = _dict if isinstance(obj, dict) else _sequence
        return c[0] + m(obj, depth) + c[1]

    def _redacted(key):
        return (
            isinstance(key, pkconst.STRING_TYPES)
            and SECRETS_RE.search(key)
            and REDACTED
        )

    def _sequence(obj, depth):
        return _iterate(
            obj,
            lambda v: _format_arg(v, depth),
            obj,
        )

    def _snip(truncated, count_of_original_elements):
        truncated = truncated[:-2]
        return (
            truncated + ", " + SNIP
            if count_of_original_elements > cfg.max_elements
            else truncated
        )

    def _string(value):
        if r"\n" in value:
            value = pkcompat.unicode_unescape(value)
        # '\n File"' is at the start of stack traces. The end of stack
        # traces are more interesting than the beginning so truncate
        # the beginning.
        if '\n  File "' in value and "Exception was printed at" not in value:
            return (
                SNIP + value[-cfg.max_string :]
                if len(value) > cfg.max_string
                else value
            )
        return value[: cfg.max_string] + (value[cfg.max_string :] and SNIP)

    try:
        f = getattr(obj, PKDEBUG_STR_FUNCTION_NAME, None)
        # only call if an instance, otherwise is unbound method
        if f and not isinstance(obj, type):
            s = f()
        else:
            if cfg.snip:
                if isinstance(obj, (dict, list, set, tuple)):
                    return _object(obj, depth)
                if isinstance(obj, pkconst.STRING_TYPES):
                    s = _string(obj)
                    # only enclose in quotes strings contained within another object
                    if depth > 0:
                        return "'" + s + "'"
                    return s
                if depth <= 0 and isinstance(obj, numbers.Number):
                    # allow format string arguments ({.2f}), which only occurs when depth=0
                    return obj
            s = str(obj)
        # Always truncates a converted string
        return _string(s)
    except Exception as e:
        _printer._err(
            "unable to format obj type={} exception={}".format(type(obj), e),
            pkdexc(),
        )
        # may get another exception later, but that would be in format.
        return obj


def _z(msg):
    """Useful for debugging this module"""
    with open("/dev/tty", "w") as f:
        f.write(str(msg) + "\n")
    return msg


cfg = pkconfig.init(
    control=(None, _cfg_control, "Pattern to match against pkdc messages"),
    max_depth=(10, int, "Maximum depth to recurse into and object when logging"),
    max_elements=(30, int, "Maximum number of elements in a dict, list, set, or tuple"),
    max_string=(8000, int, "Maximum length of an individual string"),
    output=(
        None,
        _cfg_output,
        'Where to write messages either as a "writable" or file name',
    ),
    redirect_logging=(False, bool, "Redirect Python's logging to output"),
    snip=(
        True,
        bool,
        "Redact secrets and truncate objects if true",
    ),
    want_pid_time=(False, bool, "Display pid and time in messages"),
)

if cfg:
    init()
